@@ -23,19 +23,17 @@
 
 
 #include <gst/gst.h>
+#include <gst/base/gstadapter.h>
 
 #include <FLAC/all.h>
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
-
+G_BEGIN_DECLS
 
 #define GST_TYPE_FLAC_DEC gst_flac_dec_get_type()
 #define GST_FLAC_DEC(obj) G_TYPE_CHECK_INSTANCE_CAST(obj, GST_TYPE_FLAC_DEC, GstFlacDec)
-#define GST_FLAC_DEC_CLASS(klass) G_TYPE_CHECK_CLASS_CAST(klass, GST_TYPE_FLAC_DEC, GstFlacDec)
+#define GST_FLAC_DEC_CLASS(klass) G_TYPE_CHECK_CLASS_CAST(klass, GST_TYPE_FLAC_DEC, GstFlacDecClass)
 #define GST_IS_FLAC_DEC(obj) G_TYPE_CHECK_INSTANCE_TYPE(obj, GST_TYPE_FLAC_DEC)
-#define GST_IS_FLAC_DEC_CLASS(obj) G_TYPE_CHECK_CLASS_TYPE(klass, GST_TYPE_FLAC_DEC)
+#define GST_IS_FLAC_DEC_CLASS(klass) G_TYPE_CHECK_CLASS_TYPE(klass, GST_TYPE_FLAC_DEC)
 
 typedef struct _GstFlacDec GstFlacDec;
 typedef struct _GstFlacDecClass GstFlacDecClass;
@@ -43,7 +41,11 @@ typedef struct _GstFlacDecClass GstFlacDecClass;
 struct _GstFlacDec {
   GstElement     element;
 
-  FLAC__SeekableStreamDecoder *decoder;
+  FLAC__SeekableStreamDecoder *seekable_decoder; /* for pull-based operation  */
+
+  FLAC__StreamDecoder         *stream_decoder;   /* for chain-based operation */
+  GstAdapter                  *adapter;
+  gboolean                     framed;
 
   GstPad        *sinkpad;
   GstPad        *srcpad;
@@ -67,6 +69,10 @@ struct _GstFlacDec {
   gint           depth;
   gint           width;
   gint           sample_rate;
+
+  /* from the stream info, needed for scanning */
+  guint16        min_blocksize;
+  guint16        max_blocksize;
 };
 
 struct _GstFlacDecClass {
@@ -75,9 +81,6 @@ struct _GstFlacDecClass {
 
 GType gst_flac_dec_get_type (void);
 
-
-#ifdef __cplusplus
-}
-#endif /* __cplusplus */
+G_END_DECLS
 
 #endif /* __GST_FLAC_DEC_H__ */
