@@ -32,18 +32,17 @@
 #include <gst/tag/tag.h>
 #include "gstspeexenc.h"
 
-GST_DEBUG_CATEGORY (speexenc_debug);
+GST_DEBUG_CATEGORY_STATIC (speexenc_debug);
 #define GST_CAT_DEFAULT speexenc_debug
 
 static GstPadTemplate *gst_speexenc_src_template, *gst_speexenc_sink_template;
 
 /* elementfactory information */
-GstElementDetails speexenc_details = {
-  "Speex encoder",
-  "Codec/Encoder/Audio",
-  "Encodes audio in Speex format",
-  "Wim Taymans <wim@fluendo.com>",
-};
+static const GstElementDetails speexenc_details =
+GST_ELEMENT_DETAILS ("Speex audio encoder",
+    "Codec/Encoder/Audio",
+    "Encodes audio in Speex format",
+    "Wim Taymans <wim@fluendo.com>");
 
 /* GstSpeexEnc signals and args */
 enum
@@ -98,6 +97,7 @@ gst_speexenc_get_formats (GstPad * pad)
 static void gst_speexenc_base_init (gpointer g_class);
 static void gst_speexenc_class_init (GstSpeexEncClass * klass);
 static void gst_speexenc_init (GstSpeexEnc * speexenc);
+static void gst_speexenc_finalize (GObject * object);
 
 static gboolean gst_speexenc_sinkevent (GstPad * pad, GstEvent * event);
 static GstFlowReturn gst_speexenc_chain (GstPad * pad, GstBuffer * buf);
@@ -230,9 +230,24 @@ gst_speexenc_class_init (GstSpeexEncClass * klass)
       g_param_spec_string ("last-message", "last-message",
           "The last status message", NULL, G_PARAM_READABLE));
 
-  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);
+  parent_class = g_type_class_peek_parent (klass);
 
-  gstelement_class->change_state = gst_speexenc_change_state;
+  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_speexenc_finalize);
+
+  gstelement_class->change_state =
+      GST_DEBUG_FUNCPTR (gst_speexenc_change_state);
+}
+
+static void
+gst_speexenc_finalize (GObject * object)
+{
+  GstSpeexEnc *speexenc;
+
+  speexenc = GST_SPEEXENC (object);
+
+  g_object_unref (speexenc->adapter);
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static gboolean
@@ -241,7 +256,7 @@ gst_speexenc_sink_setcaps (GstPad * pad, GstCaps * caps)
   GstSpeexEnc *speexenc;
   GstStructure *structure;
 
-  speexenc = GST_SPEEXENC (GST_PAD_PARENT (pad));
+  speexenc = GST_SPEEXENC (gst_pad_get_parent (pad));
   speexenc->setup = FALSE;
 
   structure = gst_caps_get_structure (caps, 0);
@@ -250,10 +265,9 @@ gst_speexenc_sink_setcaps (GstPad * pad, GstCaps * caps)
 
   gst_speexenc_setup (speexenc);
 
-  if (speexenc->setup)
-    return TRUE;
+  gst_object_unref (speexenc);
 
-  return FALSE;
+  return speexenc->setup;
 }
 
 static gboolean
@@ -828,27 +842,29 @@ static GstCaps *
 gst_speexenc_set_header_on_caps (GstCaps * caps, GstBuffer * buf1,
     GstBuffer * buf2)
 {
-  caps = gst_caps_make_writable (caps);
-  GstStructure *structure = gst_caps_get_structure (caps, 0);
-  GValue list = { 0 };
+  GstStructure *structure = NULL;
+  GValue array = { 0 };
   GValue value = { 0 };
 
+  caps = gst_caps_make_writable (caps);
+  structure = gst_caps_get_structure (caps, 0);
+
   /* mark buffers */
   GST_BUFFER_FLAG_SET (buf1, GST_BUFFER_FLAG_IN_CAPS);
   GST_BUFFER_FLAG_SET (buf2, GST_BUFFER_FLAG_IN_CAPS);
 
   /* put buffers in a fixed list */
-  g_value_init (&list, GST_TYPE_ARRAY);
+  g_value_init (&array, GST_TYPE_ARRAY);
   g_value_init (&value, GST_TYPE_BUFFER);
   gst_value_set_buffer (&value, buf1);
-  gst_value_list_append_value (&list, &value);
+  gst_value_array_append_value (&array, &value);
   g_value_unset (&value);
   g_value_init (&value, GST_TYPE_BUFFER);
   gst_value_set_buffer (&value, buf2);
-  gst_value_list_append_value (&list, &value);
-  gst_structure_set_value (structure, "streamheader", &list);
+  gst_value_array_append_value (&array, &value);
+  gst_structure_set_value (structure, "streamheader", &array);
   g_value_unset (&value);
-  g_value_unset (&list);
+  g_value_unset (&array);
 
   return caps;
 }
@@ -897,13 +913,8 @@ gst_speexenc_chain (GstPad * pad, GstBuffer * buf)
 
   speexenc = GST_SPEEXENC (gst_pad_get_parent (pad));
 
-  if (!speexenc->setup) {
-    gst_buffer_unref (buf);
-    GST_ELEMENT_ERROR (speexenc, CORE, NEGOTIATION, (NULL),
-        ("encoder not initialized (input is not audio?)"));
-    ret = GST_FLOW_UNEXPECTED;
-    goto error;
-  }
+  if (!speexenc->setup)
+    goto not_setup;
 
   if (!speexenc->header_sent) {
     /* Speex streams begin with two headers; the initial header (with
@@ -943,17 +954,16 @@ gst_speexenc_chain (GstPad * pad, GstBuffer * buf)
     ret = gst_speexenc_push_buffer (speexenc, buf1);
 
     if ((GST_FLOW_OK != ret) && (GST_FLOW_NOT_LINKED != ret)) {
-      gst_buffer_unref (buf1);
-      goto error;
+      /* unref buf2 as we are not going to push it anymore */
+
+      gst_buffer_unref (buf2);
+      goto done;
     }
 
     ret = gst_speexenc_push_buffer (speexenc, buf2);
 
-    if ((GST_FLOW_OK != ret) && (GST_FLOW_NOT_LINKED != ret)) {
-
-      gst_buffer_unref (buf2);
-      goto error;
-    }
+    if ((GST_FLOW_OK != ret) && (GST_FLOW_NOT_LINKED != ret))
+      goto done;
 
     speex_bits_init (&speexenc->bits);
     speex_bits_reset (&speexenc->bits);
@@ -1000,9 +1010,8 @@ gst_speexenc_chain (GstPad * pad, GstBuffer * buf)
           GST_BUFFER_OFFSET_NONE, outsize, GST_PAD_CAPS (speexenc->srcpad),
           &outbuf);
 
-      if ((GST_FLOW_OK != ret)) {
-        goto error;
-      }
+      if ((GST_FLOW_OK != ret))
+        goto done;
 
       written = speex_bits_write (&speexenc->bits,
           (gchar *) GST_BUFFER_DATA (outbuf), outsize);
@@ -1010,26 +1019,38 @@ gst_speexenc_chain (GstPad * pad, GstBuffer * buf)
       speex_bits_reset (&speexenc->bits);
 
       GST_BUFFER_TIMESTAMP (outbuf) =
-          (speexenc->frameno * frame_size -
-          speexenc->lookahead) * GST_SECOND / speexenc->rate;
-      GST_BUFFER_DURATION (outbuf) = frame_size * GST_SECOND / speexenc->rate;
-      GST_BUFFER_OFFSET (outbuf) = speexenc->bytes_out;
+          gst_util_uint64_scale_int (speexenc->frameno * frame_size -
+          speexenc->lookahead, GST_SECOND, speexenc->rate);
+      GST_BUFFER_DURATION (outbuf) = gst_util_uint64_scale_int (frame_size,
+          GST_SECOND, speexenc->rate);
+      /* set gp time and granulepos; see gst-plugins-base/ext/ogg/README */
       GST_BUFFER_OFFSET_END (outbuf) =
-          speexenc->frameno * frame_size - speexenc->lookahead;
+          ((speexenc->frameno + 1) * frame_size - speexenc->lookahead);
+      GST_BUFFER_OFFSET (outbuf) =
+          gst_util_uint64_scale_int (GST_BUFFER_OFFSET_END (outbuf), GST_SECOND,
+          speexenc->rate);
 
       ret = gst_speexenc_push_buffer (speexenc, outbuf);
 
-      if ((GST_FLOW_OK != ret) && (GST_FLOW_NOT_LINKED != ret)) {
-        gst_buffer_unref (outbuf);
-        goto error;
-      }
+      if ((GST_FLOW_OK != ret) && (GST_FLOW_NOT_LINKED != ret))
+        goto done;
     }
   }
 
-error:
-
+done:
   gst_object_unref (speexenc);
   return ret;
+
+  /* ERRORS */
+not_setup:
+  {
+    gst_buffer_unref (buf);
+    GST_ELEMENT_ERROR (speexenc, CORE, NEGOTIATION, (NULL),
+        ("encoder not initialized (input is not audio?)"));
+    ret = GST_FLOW_NOT_NEGOTIATED;
+    goto done;
+  }
+
 }
 
 
