@@ -74,6 +74,7 @@
 
 #include "gsttagdemux.h"
 
+#include <gst/base/gsttypefindhelper.h>
 #include <gst/gst-i18n-plugin.h>
 #include <string.h>
 
@@ -128,6 +129,7 @@ static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
 static void gst_tag_demux_dispose (GObject * object);
 
 static GstFlowReturn gst_tag_demux_chain (GstPad * pad, GstBuffer * buf);
+static gboolean gst_tag_demux_sink_event (GstPad * pad, GstEvent * event);
 
 static gboolean gst_tag_demux_src_activate_pull (GstPad * pad, gboolean active);
 static GstFlowReturn gst_tag_demux_read_range (GstTagDemux * tagdemux,
@@ -148,8 +150,6 @@ static GstStateChangeReturn gst_tag_demux_change_state (GstElement * element,
 static gboolean gst_tag_demux_pad_query (GstPad * pad, GstQuery * query);
 static const GstQueryType *gst_tag_demux_get_query_types (GstPad * pad);
 static gboolean gst_tag_demux_get_upstream_size (GstTagDemux * tagdemux);
-static GstCaps *gst_tag_demux_do_typefind (GstTagDemux * tagdemux,
-    GstBuffer * buffer);
 static void gst_tag_demux_send_tag_event (GstTagDemux * tagdemux);
 
 static void gst_tag_demux_base_init (gpointer g_class);
@@ -215,14 +215,17 @@ gst_tag_demux_class_init (gpointer klass, gpointer d)
 static void
 gst_tag_demux_reset (GstTagDemux * tagdemux)
 {
+  GstBuffer **buffer_p = &tagdemux->priv->collect;
+  GstCaps **caps_p = &tagdemux->priv->src_caps;
+
   tagdemux->priv->strip_start = 0;
   tagdemux->priv->strip_end = 0;
   tagdemux->priv->upstream_size = -1;
   tagdemux->priv->state = GST_TAG_DEMUX_READ_START_TAG;
   tagdemux->priv->send_tag_event = FALSE;
 
-  gst_buffer_replace (&(tagdemux->priv->collect), NULL);
-  gst_caps_replace (&(tagdemux->priv->src_caps), NULL);
+  gst_buffer_replace (buffer_p, NULL);
+  gst_caps_replace (caps_p, NULL);
 
   gst_tag_demux_remove_srcpad (tagdemux);
 
@@ -258,6 +261,8 @@ gst_tag_demux_init (GstTagDemux * demux, GstTagDemuxClass * gclass)
 
     gst_pad_set_activate_function (demux->priv->sinkpad,
         GST_DEBUG_FUNCPTR (gst_tag_demux_sink_activate));
+    gst_pad_set_event_function (demux->priv->sinkpad,
+        GST_DEBUG_FUNCPTR (gst_tag_demux_sink_event));
     gst_pad_set_chain_function (demux->priv->sinkpad,
         GST_DEBUG_FUNCPTR (gst_tag_demux_chain));
     gst_element_add_pad (GST_ELEMENT (demux), demux->priv->sinkpad);
@@ -348,6 +353,10 @@ gst_tag_demux_remove_srcpad (GstTagDemux * demux)
   return res;
 };
 
+/* will return FALSE if buffer is beyond end of data; will return TRUE
+ * if buffer was trimmed successfully or didn't need trimming, but may
+ * also return TRUE and set *buf_ref to NULL if the buffer was before
+ * the start of the data */
 static gboolean
 gst_tag_demux_trim_buffer (GstTagDemux * tagdemux, GstBuffer ** buf_ref)
 {
@@ -372,7 +381,7 @@ gst_tag_demux_trim_buffer (GstTagDemux * tagdemux, GstBuffer ** buf_ref)
 
       if (out_offset >= v1tag_offset) {
         GST_DEBUG_OBJECT (tagdemux, "Buffer is past the end of the data");
-        goto no_out_buffer;
+        goto no_out_buffer_end;
       }
 
       if (out_offset + out_size > v1tag_offset) {
@@ -387,7 +396,7 @@ gst_tag_demux_trim_buffer (GstTagDemux * tagdemux, GstBuffer ** buf_ref)
     if (out_offset <= tagdemux->priv->strip_start) {
       if (out_offset + out_size <= tagdemux->priv->strip_start) {
         GST_DEBUG_OBJECT (tagdemux, "Buffer is before the start of the data");
-        goto no_out_buffer;
+        goto no_out_buffer_start;
       }
 
       trim_start = tagdemux->priv->strip_start - out_offset;
@@ -426,10 +435,18 @@ gst_tag_demux_trim_buffer (GstTagDemux * tagdemux, GstBuffer ** buf_ref)
 
   return TRUE;
 
-no_out_buffer:
-  gst_buffer_unref (buf);
-  *buf_ref = NULL;
-  return TRUE;
+no_out_buffer_end:
+  {
+    gst_buffer_unref (buf);
+    *buf_ref = NULL;
+    return FALSE;
+  }
+no_out_buffer_start:
+  {
+    gst_buffer_unref (buf);
+    *buf_ref = NULL;
+    return TRUE;
+  }
 }
 
 static void
@@ -537,10 +554,12 @@ gst_tag_demux_chain (GstPad * pad, GstBuffer * buf)
         break;
       /* Fall-through */
     case GST_TAG_DEMUX_TYPEFINDING:{
-      GstCaps *caps;
+      GstTypeFindProbability probability = 0;
       GstBuffer *typefind_buf = NULL;
+      GstCaps *caps;
 
-      if (GST_BUFFER_SIZE (demux->priv->collect) < TYPE_FIND_MIN_SIZE)
+      if (GST_BUFFER_SIZE (demux->priv->collect) <
+          TYPE_FIND_MIN_SIZE + demux->priv->strip_start)
         break;                  /* Go get more data first */
 
       GST_DEBUG_OBJECT (demux, "Typefinding with size %d",
@@ -550,9 +569,13 @@ gst_tag_demux_chain (GstPad * pad, GstBuffer * buf)
       typefind_buf = demux->priv->collect;
       gst_buffer_ref (typefind_buf);
       if (!gst_tag_demux_trim_buffer (demux, &typefind_buf))
-        return GST_FLOW_ERROR;
+        return GST_FLOW_UNEXPECTED;
 
-      caps = gst_tag_demux_do_typefind (demux, typefind_buf);
+      if (typefind_buf == NULL)
+        break;                  /* Still need more data */
+
+      caps = gst_type_find_helper_for_buffer (GST_OBJECT (demux),
+          typefind_buf, &probability);
 
       if (caps == NULL) {
         if (GST_BUFFER_SIZE (typefind_buf) < TYPE_FIND_MAX_SIZE) {
@@ -570,6 +593,9 @@ gst_tag_demux_chain (GstPad * pad, GstBuffer * buf)
       }
       gst_buffer_unref (typefind_buf);
 
+      GST_DEBUG_OBJECT (demux, "Found type %" GST_PTR_FORMAT " with a "
+          "probability of %u", caps, probability);
+
       if (!gst_tag_demux_add_srcpad (demux, caps)) {
         GST_DEBUG_OBJECT (demux, "Failed to add srcpad");
         gst_caps_unref (caps);
@@ -595,7 +621,7 @@ gst_tag_demux_chain (GstPad * pad, GstBuffer * buf)
         outbuf = demux->priv->collect;
         demux->priv->collect = NULL;
         if (!gst_tag_demux_trim_buffer (demux, &outbuf))
-          return GST_FLOW_ERROR;
+          return GST_FLOW_UNEXPECTED;
       }
       if (outbuf) {
         if (G_UNLIKELY (demux->priv->srcpad == NULL)) {
@@ -606,6 +632,11 @@ gst_tag_demux_chain (GstPad * pad, GstBuffer * buf)
         GST_DEBUG_OBJECT (demux, "Pushing buffer %p", outbuf);
         /* gst_util_dump_mem (GST_BUFFER_DATA (outbuf),
            GST_BUFFER_SIZE (outbuf)); */
+
+        /* Ensure the caps are set correctly */
+        outbuf = gst_buffer_make_metadata_writable (outbuf);
+        gst_buffer_set_caps (outbuf, GST_PAD_CAPS (demux->priv->srcpad));
+
         return gst_pad_push (demux->priv->srcpad, outbuf);
       }
     }
@@ -618,39 +649,49 @@ error:
   return GST_FLOW_ERROR;
 }
 
+static gboolean
+gst_tag_demux_sink_event (GstPad * pad, GstEvent * event)
+{
+  GstTagDemux *demux;
+  gboolean ret;
+
+  demux = GST_TAG_DEMUX (gst_pad_get_parent (pad));
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_EOS:
+      if (demux->priv->srcpad == NULL) {
+        GST_WARNING_OBJECT (demux, "EOS before we found a type");
+        GST_ELEMENT_ERROR (demux, STREAM, TYPE_NOT_FOUND, (NULL), (NULL));
+      }
+      ret = gst_pad_event_default (pad, event);
+      break;
+    default:
+      ret = gst_pad_event_default (pad, event);
+      break;
+  }
+
+  gst_object_unref (demux);
+  return ret;
+}
+
 static gboolean
 gst_tag_demux_get_upstream_size (GstTagDemux * tagdemux)
 {
-  GstQuery *query;
-  GstPad *peer = NULL;
   GstFormat format;
-  gint64 result;
-  gboolean res = FALSE;
+  gint64 len;
 
   /* Short-cut if we already queried upstream */
   if (tagdemux->priv->upstream_size > 0)
     return TRUE;
 
-  if ((peer = gst_pad_get_peer (tagdemux->priv->sinkpad)) == NULL)
+  format = GST_FORMAT_BYTES;
+  if (!gst_pad_query_peer_duration (tagdemux->priv->sinkpad, &format, &len) ||
+      len <= 0) {
     return FALSE;
+  }
 
-  query = gst_query_new_duration (GST_FORMAT_BYTES);
-  gst_query_set_duration (query, GST_FORMAT_BYTES, -1);
-
-  if (!gst_pad_query (peer, query))
-    goto out;
-
-  gst_query_parse_duration (query, &format, &result);
-
-  if (format != GST_FORMAT_BYTES || result == -1)
-    goto out;
-
-  tagdemux->priv->upstream_size = result;
-  res = TRUE;
-
-out:
-  gst_object_unref (peer);
-  return res;
+  tagdemux->priv->upstream_size = len;
+  return TRUE;
 }
 
 static gboolean
@@ -1003,11 +1044,10 @@ done:
 static gboolean
 gst_tag_demux_sink_activate (GstPad * sinkpad)
 {
+  GstTypeFindProbability probability = 0;
   GstTagDemux *demux = GST_TAG_DEMUX (GST_PAD_PARENT (sinkpad));
   gboolean ret = FALSE;
-  GstBuffer *buf = NULL;
   GstCaps *caps = NULL;
-  GstFlowReturn flow_ret;
 
   /* 1: */
   /* If we can activate pull_range upstream, then read any end and start
@@ -1048,17 +1088,14 @@ gst_tag_demux_sink_activate (GstPad * sinkpad)
     demux->priv->send_tag_event = TRUE;
   }
 
-  flow_ret = gst_tag_demux_read_range (demux, 0, TYPE_FIND_MAX_SIZE, &buf);
-  if (flow_ret != GST_FLOW_OK) {
-    GST_DEBUG_OBJECT (demux, "Could not read data from start of file ret=%s",
-        gst_flow_get_name (flow_ret));
-    goto done_activate;
-  }
+  /* 3 - Do typefinding on data */
+  caps = gst_type_find_helper_get_range (GST_OBJECT (demux),
+      (GstTypeFindHelperGetRangeFunction) gst_tag_demux_read_range,
+      demux->priv->upstream_size
+      - (demux->priv->strip_start + demux->priv->strip_end), &probability);
 
-  /* gst_util_dump_mem (GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf)); */
-  caps = gst_tag_demux_do_typefind (demux, buf);
-  gst_buffer_unref (buf);
-  buf = NULL;
+  GST_DEBUG_OBJECT (demux, "Found type %" GST_PTR_FORMAT " with a "
+      "probability of %u", caps, probability);
 
   /* 4 - Deactivate pull mode */
   if (!gst_pad_activate_pull (sinkpad, FALSE)) {
@@ -1095,8 +1132,6 @@ gst_tag_demux_sink_activate (GstPad * sinkpad)
   }
 
 done_activate:
-  if (buf)
-    gst_buffer_unref (buf);
 
   return ret;
 }
@@ -1135,26 +1170,35 @@ gst_tag_demux_read_range (GstTagDemux * demux,
   if (!gst_tag_demux_get_upstream_size (demux))
     return GST_FLOW_ERROR;
 
-  if (in_offset + length >= demux->priv->upstream_size - demux->priv->strip_end)
+  if (in_offset + length >= demux->priv->upstream_size - demux->priv->strip_end) {
+    if (in_offset + demux->priv->strip_end >= demux->priv->upstream_size)
+      return GST_FLOW_UNEXPECTED;
     in_length = demux->priv->upstream_size - demux->priv->strip_end - in_offset;
-  else
+  } else {
     in_length = length;
+  }
 
   ret = gst_pad_pull_range (demux->priv->sinkpad, in_offset, in_length, buffer);
 
   if (ret == GST_FLOW_OK && *buffer) {
     if (!gst_tag_demux_trim_buffer (demux, buffer))
-      goto error;
+      goto read_beyond_end;
+
+    /* this should only happen in streaming mode */
+    g_assert (*buffer != NULL);
   }
 
   return ret;
 
-error:
-  if (*buffer != NULL) {
-    gst_buffer_unref (buffer);
-    *buffer = NULL;
+read_beyond_end:
+  {
+    GST_DEBUG_OBJECT (demux, "attempted read beyond end of file");
+    if (*buffer != NULL) {
+      gst_buffer_unref (buffer);
+      *buffer = NULL;
+    }
+    return GST_FLOW_UNEXPECTED;
   }
-  return GST_FLOW_ERROR;
 }
 
 static GstFlowReturn
@@ -1254,74 +1298,6 @@ gst_tag_demux_get_query_types (GstPad * pad)
   return types;
 }
 
-typedef struct
-{
-  guint best_probability;
-  GstCaps *caps;
-  GstBuffer *buffer;
-} SimpleTypeFind;
-
-static guint8 *
-simple_find_peek (gpointer data, gint64 offset, guint size)
-{
-  SimpleTypeFind *find = (SimpleTypeFind *) data;
-
-  if (offset < 0)
-    return NULL;
-
-  if (GST_BUFFER_SIZE (find->buffer) >= offset + size) {
-    return GST_BUFFER_DATA (find->buffer) + offset;
-  }
-  return NULL;
-}
-static void
-simple_find_suggest (gpointer data, guint probability, const GstCaps * caps)
-{
-  SimpleTypeFind *find = (SimpleTypeFind *) data;
-
-  if (probability > find->best_probability) {
-    GstCaps *copy = gst_caps_copy (caps);
-
-    gst_caps_replace (&find->caps, copy);
-    gst_caps_unref (copy);
-    find->best_probability = probability;
-  }
-}
-
-static GstCaps *
-gst_tag_demux_do_typefind (GstTagDemux * tagdemux, GstBuffer * buffer)
-{
-  GList *walk, *type_list;
-  SimpleTypeFind find;
-  GstTypeFind gst_find;
-
-  walk = type_list = gst_type_find_factory_get_list ();
-
-  find.buffer = buffer;
-  find.best_probability = 0;
-  find.caps = NULL;
-  gst_find.data = &find;
-  gst_find.peek = simple_find_peek;
-  gst_find.get_length = NULL;
-  gst_find.suggest = simple_find_suggest;
-  while (walk) {
-    GstTypeFindFactory *factory = GST_TYPE_FIND_FACTORY (walk->data);
-
-    gst_type_find_factory_call_function (factory, &gst_find);
-    if (find.best_probability >= GST_TYPE_FIND_MAXIMUM)
-      break;
-    walk = g_list_next (walk);
-  }
-  gst_plugin_feature_list_free (type_list);
-  if (find.best_probability > 0) {
-    GST_DEBUG ("Found caps %" GST_PTR_FORMAT " with buf size %u", find.caps,
-        GST_BUFFER_SIZE (buffer));
-    return find.caps;
-  }
-
-  return NULL;
-}
-
 static void
 gst_tag_demux_send_tag_event (GstTagDemux * demux)
 {
