@@ -33,18 +33,11 @@
 GST_DEBUG_CATEGORY_STATIC (matroskamux_debug);
 #define GST_CAT_DEFAULT matroskamux_debug
 
-enum
-{
-  /* FILL ME */
-  LAST_SIGNAL
-};
-
 enum
 {
   ARG_0,
   ARG_WRITING_APP,
   ARG_MATROSKA_VERSION
-      /* FILL ME */
 };
 
 static GstStaticPadTemplate src_templ = GST_STATIC_PAD_TEMPLATE ("src",
@@ -72,10 +65,17 @@ static GstStaticPadTemplate videosink_templ =
         COMMON_VIDEO_CAPS "; "
         "video/x-xvid, "
         COMMON_VIDEO_CAPS "; "
+        "video/x-huffyuv, "
+        COMMON_VIDEO_CAPS "; "
+        "video/x-dv, "
+        COMMON_VIDEO_CAPS "; "
+        "video/x-h263, "
+        COMMON_VIDEO_CAPS "; "
         "video/x-msmpeg, "
         COMMON_VIDEO_CAPS "; "
         "image/jpeg, "
         COMMON_VIDEO_CAPS "; "
+        "video/x-theora; "
         "video/x-raw-yuv, "
         "format = (fourcc) { YUY2, I420 }, " COMMON_VIDEO_CAPS)
     );
@@ -121,15 +121,8 @@ GST_STATIC_PAD_TEMPLATE ("subtitle_%d",
 
 static GArray *used_uids;
 
-static void
-_do_init (GType matroskamux_type)
-{
-  GST_DEBUG_CATEGORY_INIT (matroskamux_debug, "matroskamux", 0,
-      "Matroska muxer");
-}
-
-GST_BOILERPLATE_FULL (GstMatroskaMux, gst_matroska_mux, GstElement,
-    GST_TYPE_ELEMENT, _do_init);
+GST_BOILERPLATE (GstMatroskaMux, gst_matroska_mux, GstElement,
+    GST_TYPE_ELEMENT);
 
 /* Matroska muxer destructor */
 static void gst_matroska_mux_finalize (GObject * object);
@@ -160,18 +153,20 @@ static void gst_matroska_mux_reset (GstElement * element);
 /* uid generation */
 static guint32 gst_matroska_mux_create_uid ();
 
-/*static guint gst_matroska_mux_signals[LAST_SIGNAL] = { 0 };*/
+static gboolean theora_streamheader_to_codecdata (const GValue * streamheader,
+    GstMatroskaTrackContext * context);
+static gboolean vorbis_streamheader_to_codecdata (const GValue * streamheader,
+    GstMatroskaTrackContext * context);
 
 static void
 gst_matroska_mux_base_init (gpointer g_class)
 {
   GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);
-  static GstElementDetails gst_matroska_mux_details = {
-    "Matroska muxer",
-    "Codec/Muxer",
-    "Muxes video/audio/subtitle streams into a matroska stream",
-    "Ronald Bultje <rbultje@ronald.bitfreak.net>"
-  };
+  static const GstElementDetails gst_matroska_mux_details =
+      GST_ELEMENT_DETAILS ("Matroska muxer",
+      "Codec/Muxer",
+      "Muxes video/audio/subtitle streams into a matroska stream",
+      "Ronald Bultje <rbultje@ronald.bitfreak.net>");
 
   gst_element_class_add_pad_template (element_class,
       gst_static_pad_template_get (&videosink_templ));
@@ -182,6 +177,9 @@ gst_matroska_mux_base_init (gpointer g_class)
   gst_element_class_add_pad_template (element_class,
       gst_static_pad_template_get (&src_templ));
   gst_element_class_set_details (element_class, &gst_matroska_mux_details);
+
+  GST_DEBUG_CATEGORY_INIT (matroskamux_debug, "matroskamux", 0,
+      "Matroska muxer");
 }
 
 static void
@@ -232,7 +230,8 @@ gst_matroska_mux_init (GstMatroskaMux * mux, GstMatroskaMuxClass * g_class)
 
   mux->collect = gst_collect_pads_new ();
   gst_collect_pads_set_function (mux->collect,
-      (GstCollectPadsFunction) gst_matroska_mux_collected, mux);
+      (GstCollectPadsFunction) GST_DEBUG_FUNCPTR (gst_matroska_mux_collected),
+      mux);
 
   mux->ebml_write = gst_ebml_write_new (mux->srcpad);
 
@@ -421,12 +420,15 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
 {
   GstMatroskaTrackContext *context = NULL;
   GstMatroskaTrackVideoContext *videocontext;
+  GstMatroskaMux *mux;
   GstMatroskaPad *collect_pad;
   const gchar *mimetype;
   gint width, height, pixel_width, pixel_height;
   const GValue *framerate;
   GstStructure *structure;
 
+  mux = GST_MATROSKA_MUX (GST_PAD_PARENT (pad));
+
   /* find context */
   collect_pad = (GstMatroskaPad *) gst_pad_get_element_private (pad);
   g_assert (collect_pad);
@@ -440,6 +442,11 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
 
   mimetype = gst_structure_get_name (structure);
 
+  if (!strcmp (mimetype, "video/x-theora")) {
+    /* we'll extract the details later from the theora identification header */
+    goto skip_details;
+  }
+
   /* get general properties */
   gst_structure_get_int (structure, "width", &width);
   gst_structure_get_int (structure, "height", &height);
@@ -470,6 +477,8 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
     videocontext->display_height = 0;
   }
 
+skip_details:
+
   videocontext->asr_mode = GST_MATROSKA_ASPECT_RATIO_MODE_FREE;
   videocontext->eye_mode = GST_MATROSKA_EYE_MODE_MONO;
   videocontext->fourcc = 0;
@@ -484,12 +493,17 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
     context->codec_id = g_strdup (GST_MATROSKA_CODEC_ID_VIDEO_MJPEG);
 
     return TRUE;
-  } else if (!strcmp (mimetype, "video/x-divx")) {
-    gint divxversion;
+  } else if (!strcmp (mimetype, "video/x-xvid") /* MS/VfW compatibility cases */
+      ||!strcmp (mimetype, "video/x-huffyuv")
+      || !strcmp (mimetype, "video/x-divx")
+      || !strcmp (mimetype, "video/x-dv")
+      || !strcmp (mimetype, "video/x-h263")) {
     BITMAPINFOHEADER *bih;
+    const GValue *codec_data;
+    gint size = sizeof (BITMAPINFOHEADER);
 
-    bih = (BITMAPINFOHEADER *) g_malloc0 (sizeof (BITMAPINFOHEADER));
-    GST_WRITE_UINT32_LE (&bih->bi_size, sizeof (BITMAPINFOHEADER));
+    bih = g_new0 (BITMAPINFOHEADER, 1);
+    GST_WRITE_UINT32_LE (&bih->bi_size, size);
     GST_WRITE_UINT32_LE (&bih->bi_width, videocontext->pixel_width);
     GST_WRITE_UINT32_LE (&bih->bi_height, videocontext->pixel_height);
     GST_WRITE_UINT16_LE (&bih->bi_planes, (guint16) 1);
@@ -497,40 +511,47 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
     GST_WRITE_UINT32_LE (&bih->bi_size_image, videocontext->pixel_width *
         videocontext->pixel_height * 3);
 
-    gst_structure_get_int (structure, "divxversion", &divxversion);
-    switch (divxversion) {
-      case 3:
-        GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DIV3"));
-        break;
-      case 4:
-        GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DIVX"));
-        break;
-      case 5:
-        GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DX50"));
-        break;
+    if (!strcmp (mimetype, "video/x-xvid"))
+      GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("XVID"));
+    else if (!strcmp (mimetype, "video/x-huffyuv"))
+      GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("HFYU"));
+    else if (!strcmp (mimetype, "video/x-dv"))
+      GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DVSD"));
+    else if (!strcmp (mimetype, "video/x-h263"))
+      GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("H263"));
+    else if (!strcmp (mimetype, "video/x-divx")) {
+      gint divxversion;
+
+      gst_structure_get_int (structure, "divxversion", &divxversion);
+      switch (divxversion) {
+        case 3:
+          GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DIV3"));
+          break;
+        case 4:
+          GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DIVX"));
+          break;
+        case 5:
+          GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("DX50"));
+          break;
+      }
+    }
+
+    /* process codec private/initialization data, if any */
+    codec_data = gst_structure_get_value (structure, "codec_data");
+    if (codec_data) {
+      GstBuffer *codec_data_buf;
+
+      codec_data_buf = g_value_peek_pointer (codec_data);
+      size += GST_BUFFER_SIZE (codec_data_buf);
+      bih = g_realloc (bih, size);
+      GST_WRITE_UINT32_LE (&bih->bi_size, size);
+      memcpy ((guint8 *) bih + sizeof (BITMAPINFOHEADER),
+          GST_BUFFER_DATA (codec_data_buf), GST_BUFFER_SIZE (codec_data_buf));
     }
 
     context->codec_id = g_strdup (GST_MATROSKA_CODEC_ID_VIDEO_VFW_FOURCC);
     context->codec_priv = (gpointer) bih;
-    context->codec_priv_size = sizeof (BITMAPINFOHEADER);
-
-    return TRUE;
-  } else if (!strcmp (mimetype, "video/x-xvid")) {
-    BITMAPINFOHEADER *bih;
-
-    bih = (BITMAPINFOHEADER *) g_malloc0 (sizeof (BITMAPINFOHEADER));
-    GST_WRITE_UINT32_LE (&bih->bi_size, sizeof (BITMAPINFOHEADER));
-    GST_WRITE_UINT32_LE (&bih->bi_width, videocontext->pixel_width);
-    GST_WRITE_UINT32_LE (&bih->bi_height, videocontext->pixel_height);
-    GST_WRITE_UINT16_LE (&bih->bi_planes, (guint16) 1);
-    GST_WRITE_UINT16_LE (&bih->bi_bit_count, (guint16) 24);
-    GST_WRITE_UINT32_LE (&bih->bi_compression, GST_STR_FOURCC ("XVID"));
-    GST_WRITE_UINT32_LE (&bih->bi_size_image, videocontext->pixel_width *
-        videocontext->pixel_height * 3);
-
-    context->codec_id = g_strdup (GST_MATROSKA_CODEC_ID_VIDEO_VFW_FOURCC);
-    context->codec_priv = (gpointer) bih;
-    context->codec_priv_size = sizeof (BITMAPINFOHEADER);
+    context->codec_priv_size = size;
 
     return TRUE;
   } else if (!strcmp (mimetype, "video/x-h264")) {
@@ -563,6 +584,24 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
       context->codec_priv_size = priv_data_size;
     }
 
+    return TRUE;
+  } else if (!strcmp (mimetype, "video/x-theora")) {
+    const GValue *streamheader;
+
+    context->codec_id = g_strdup (GST_MATROSKA_CODEC_ID_VIDEO_THEORA);
+
+    if (context->codec_priv != NULL) {
+      g_free (context->codec_priv);
+      context->codec_priv = NULL;
+      context->codec_priv_size = 0;
+    }
+
+    streamheader = gst_structure_get_value (structure, "streamheader");
+    if (!theora_streamheader_to_codecdata (streamheader, context)) {
+      GST_ELEMENT_ERROR (mux, STREAM, MUX, (NULL),
+          ("theora stream headers missing or malformed"));
+      return FALSE;
+    }
     return TRUE;
   } else if (!strcmp (mimetype, "video/mpeg")) {
     gint mpegversion;
@@ -592,6 +631,182 @@ gst_matroska_mux_video_pad_setcaps (GstPad * pad, GstCaps * caps)
   return FALSE;
 }
 
+static gboolean
+xiph3_streamheader_to_codecdata (const GValue * streamheader,
+    GstMatroskaTrackContext * context, GstBuffer ** p_buf0)
+{
+  GstBuffer *buf[3];
+  GArray *bufarr;
+  guint8 *priv_data;
+  guint i, offset, priv_data_size;
+
+  if (streamheader == NULL)
+    goto no_stream_headers;
+
+  if (G_VALUE_TYPE (streamheader) != GST_TYPE_ARRAY)
+    goto wrong_type;
+
+  bufarr = g_value_peek_pointer (streamheader);
+  if (bufarr->len != 3)
+    goto wrong_count;
+
+  context->xiph_headers_to_skip = bufarr->len;
+
+  for (i = 0; i < 3; i++) {
+    GValue *bufval = &g_array_index (bufarr, GValue, i);
+
+    if (G_VALUE_TYPE (bufval) != GST_TYPE_BUFFER)
+      goto wrong_content_type;
+
+    buf[i] = g_value_peek_pointer (bufval);
+  }
+
+  priv_data_size = 1;
+  priv_data_size += GST_BUFFER_SIZE (buf[0]) / 0xff + 1;
+  priv_data_size += GST_BUFFER_SIZE (buf[1]) / 0xff + 1;
+
+  for (i = 0; i < 3; ++i) {
+    priv_data_size += GST_BUFFER_SIZE (buf[i]);
+  }
+
+  priv_data = g_malloc0 (priv_data_size);
+
+  priv_data[0] = 2;
+  offset = 1;
+
+  for (i = 0; i < GST_BUFFER_SIZE (buf[0]) / 0xff; ++i) {
+    priv_data[offset++] = 0xff;
+  }
+  priv_data[offset++] = GST_BUFFER_SIZE (buf[0]) % 0xff;
+
+  for (i = 0; i < GST_BUFFER_SIZE (buf[1]) / 0xff; ++i) {
+    priv_data[offset++] = 0xff;
+  }
+  priv_data[offset++] = GST_BUFFER_SIZE (buf[1]) % 0xff;
+
+  for (i = 0; i < 3; ++i) {
+    memcpy (priv_data + offset, GST_BUFFER_DATA (buf[i]),
+        GST_BUFFER_SIZE (buf[i]));
+    offset += GST_BUFFER_SIZE (buf[i]);
+  }
+
+  context->codec_priv = priv_data;
+  context->codec_priv_size = priv_data_size;
+
+  if (p_buf0)
+    *p_buf0 = gst_buffer_ref (buf[0]);
+
+  return TRUE;
+
+/* ERRORS */
+no_stream_headers:
+  {
+    GST_WARNING ("required streamheaders missing in sink caps!");
+    return FALSE;
+  }
+wrong_type:
+  {
+    GST_WARNING ("streamheaders are not a GST_TYPE_ARRAY, but a %s",
+        G_VALUE_TYPE_NAME (streamheader));
+    return FALSE;
+  }
+wrong_count:
+  {
+    GST_WARNING ("got %u streamheaders, not 3 as expected", bufarr->len);
+    return FALSE;
+  }
+wrong_content_type:
+  {
+    GST_WARNING ("streamheaders array does not contain GstBuffers");
+    return FALSE;
+  }
+}
+
+static gboolean
+vorbis_streamheader_to_codecdata (const GValue * streamheader,
+    GstMatroskaTrackContext * context)
+{
+  GstBuffer *buf0 = NULL;
+
+  if (!xiph3_streamheader_to_codecdata (streamheader, context, &buf0))
+    return FALSE;
+
+  if (buf0 == NULL || GST_BUFFER_SIZE (buf0) < 1 + 6 + 4) {
+    GST_WARNING ("First vorbis header too small, ignoring");
+  } else {
+    if (memcmp (GST_BUFFER_DATA (buf0) + 1, "vorbis", 6) == 0) {
+      GstMatroskaTrackAudioContext *audiocontext;
+      guint8 *hdr;
+
+      hdr = GST_BUFFER_DATA (buf0) + 1 + 6 + 4;
+      audiocontext = (GstMatroskaTrackAudioContext *) context;
+      audiocontext->channels = GST_READ_UINT8 (hdr);
+      audiocontext->samplerate = GST_READ_UINT32_LE (hdr + 1);
+    }
+  }
+
+  if (buf0)
+    gst_buffer_unref (buf0);
+
+  return TRUE;
+}
+
+static gboolean
+theora_streamheader_to_codecdata (const GValue * streamheader,
+    GstMatroskaTrackContext * context)
+{
+  GstBuffer *buf0 = NULL;
+
+  if (!xiph3_streamheader_to_codecdata (streamheader, context, &buf0))
+    return FALSE;
+
+  if (buf0 == NULL || GST_BUFFER_SIZE (buf0) < 1 + 6 + 26) {
+    GST_WARNING ("First theora header too small, ignoring");
+  } else if (memcmp (GST_BUFFER_DATA (buf0), "\200theora\003\002", 9) != 0) {
+    GST_WARNING ("First header not a theora identification header, ignoring");
+  } else {
+    GstMatroskaTrackVideoContext *videocontext;
+    guint fps_num, fps_denom, par_num, par_denom;
+    guint8 *hdr;
+
+    hdr = GST_BUFFER_DATA (buf0) + 1 + 6 + 3 + 2 + 2;
+
+    videocontext = (GstMatroskaTrackVideoContext *) context;
+    videocontext->pixel_width = GST_READ_UINT32_BE (hdr) >> 8;
+    videocontext->pixel_height = GST_READ_UINT32_BE (hdr + 3) >> 8;
+    hdr += 3 + 3 + 1 + 1;
+    fps_num = GST_READ_UINT32_BE (hdr);
+    fps_denom = GST_READ_UINT32_BE (hdr + 4);
+    context->default_duration = gst_util_uint64_scale_int (GST_SECOND,
+        fps_denom, fps_num);
+    hdr += 4 + 4;
+    par_num = GST_READ_UINT32_BE (hdr) >> 8;
+    par_denom = GST_READ_UINT32_BE (hdr + 3) >> 8;
+    if (par_num > 0 && par_num > 0) {
+      if (par_num > par_denom) {
+        videocontext->display_width =
+            videocontext->pixel_width * par_num / par_denom;
+        videocontext->display_height = videocontext->pixel_height;
+      } else if (par_num < par_denom) {
+        videocontext->display_width = videocontext->pixel_width;
+        videocontext->display_height =
+            videocontext->pixel_height * par_denom / par_num;
+      } else {
+        videocontext->display_width = 0;
+        videocontext->display_height = 0;
+      }
+    } else {
+      videocontext->display_width = 0;
+      videocontext->display_height = 0;
+    }
+    hdr += 3 + 3;
+  }
+
+  if (buf0)
+    gst_buffer_unref (buf0);
+
+  return TRUE;
+}
 
 /**
  * gst_matroska_mux_audio_pad_setcaps:
@@ -607,11 +822,14 @@ gst_matroska_mux_audio_pad_setcaps (GstPad * pad, GstCaps * caps)
 {
   GstMatroskaTrackContext *context = NULL;
   GstMatroskaTrackAudioContext *audiocontext;
+  GstMatroskaMux *mux;
   GstMatroskaPad *collect_pad;
   const gchar *mimetype;
   gint samplerate = 0, channels = 0;
   GstStructure *structure;
 
+  mux = GST_MATROSKA_MUX (GST_PAD_PARENT (pad));
+
   /* find context */
   collect_pad = (GstMatroskaPad *) gst_pad_get_element_private (pad);
   g_assert (collect_pad);
@@ -698,8 +916,6 @@ gst_matroska_mux_audio_pad_setcaps (GstPad * pad, GstCaps * caps)
     /* FIXME: endianness is undefined */
   } else if (!strcmp (mimetype, "audio/x-vorbis")) {
     const GValue *streamheader;
-    guint8 *priv_data = NULL;
-    guint priv_data_size = 0;
 
     context->codec_id = g_strdup (GST_MATROSKA_CODEC_ID_AUDIO_VORBIS);
 
@@ -708,81 +924,14 @@ gst_matroska_mux_audio_pad_setcaps (GstPad * pad, GstCaps * caps)
       context->codec_priv = NULL;
       context->codec_priv_size = 0;
     }
+
     streamheader = gst_structure_get_value (structure, "streamheader");
-
-    if (streamheader != NULL) {
-      if (G_VALUE_TYPE (streamheader) == GST_TYPE_ARRAY) {
-        GArray *bufarr = g_value_peek_pointer (streamheader);
-        gint i;
-        gint offset;
-
-        if (bufarr->len == 3) {
-          GstBuffer *buf[3];
-
-          for (i = 0; i < bufarr->len; i++) {
-            GValue *bufval = &g_array_index (bufarr, GValue, i);
-
-            if (G_VALUE_TYPE (bufval) == GST_TYPE_BUFFER) {
-              buf[i] = g_value_peek_pointer (bufval);
-            }
-          }
-
-          priv_data_size = 1;
-          priv_data_size += GST_BUFFER_SIZE (buf[0]) / 0xff + 1;
-          priv_data_size += GST_BUFFER_SIZE (buf[1]) / 0xff + 1;
-
-          for (i = 0; i < 3; ++i) {
-            priv_data_size += GST_BUFFER_SIZE (buf[i]);
-          }
-
-          priv_data = g_malloc0 (priv_data_size);
-
-          priv_data[0] = 2;
-          offset = 1;
-
-          for (i = 0; i < GST_BUFFER_SIZE (buf[0]) / 0xff; ++i) {
-            priv_data[offset++] = 0xff;
-          }
-          priv_data[offset++] = GST_BUFFER_SIZE (buf[0]) % 0xff;
-
-          for (i = 0; i < GST_BUFFER_SIZE (buf[1]) / 0xff; ++i) {
-            priv_data[offset++] = 0xff;
-          }
-          priv_data[offset++] = GST_BUFFER_SIZE (buf[1]) % 0xff;
-
-          for (i = 0; i < 3; ++i) {
-            memcpy (priv_data + offset, GST_BUFFER_DATA (buf[i]),
-                GST_BUFFER_SIZE (buf[i]));
-            offset += GST_BUFFER_SIZE (buf[i]);
-          }
-
-          if (GST_BUFFER_SIZE (buf[0]) < 1 + 6 + 4) {
-            GST_WARNING ("First vorbis header too small, ignoring");
-          } else {
-            if (memcmp (GST_BUFFER_DATA (buf[0]) + 1, "vorbis", 6) == 0) {
-              guint8 *hdr = GST_BUFFER_DATA (buf[0]) + 1 + 6 + 4;
-
-              audiocontext->channels = GST_READ_UINT8 (hdr);
-              audiocontext->samplerate = GST_READ_UINT32_LE (hdr + 1);
-            }
-          }
-        } else {
-          GST_WARNING ("Vorbis header does not contain "
-              "three buffers (found %d buffers), Ignoring.", bufarr->len);
-        }
-      }
+    if (!vorbis_streamheader_to_codecdata (streamheader, context)) {
+      GST_ELEMENT_ERROR (mux, STREAM, MUX, (NULL),
+          ("vorbis stream headers missing or malformed"));
+      return FALSE;
     }
-
-    if (priv_data == NULL) {
-      GST_WARNING ("Could not write Vorbis header into codec private data. "
-          "You will probably not be able to play the stream.");
-    }
-
-    context->codec_priv_size = priv_data_size;
-    context->codec_priv = (gpointer) priv_data;
-
     return TRUE;
-
   } else if (!strcmp (mimetype, "audio/x-ac3")) {
     context->codec_id = g_strdup (GST_MATROSKA_CODEC_ID_AUDIO_AC3);
 
@@ -1066,9 +1215,9 @@ gst_matroska_mux_start (GstMatroskaMux * mux)
     gst_object_unref (peerpad);
   }
   gst_ebml_write_float (ebml, GST_MATROSKA_ID_DURATION,
-      duration / mux->time_scale);
+      duration / gst_guint64_to_gdouble (mux->time_scale));
   gst_ebml_write_utf8 (ebml, GST_MATROSKA_ID_MUXINGAPP,
-      "GStreamer plugin version " GST_PLUGINS_GOOD_VERSION);
+      "GStreamer plugin version " PACKAGE_VERSION);
   if (mux->writing_app && mux->writing_app[0]) {
     gst_ebml_write_utf8 (ebml, GST_MATROSKA_ID_WRITINGAPP, mux->writing_app);
   }
@@ -1232,7 +1381,7 @@ gst_matroska_mux_finish (GstMatroskaMux * mux)
     pos = mux->ebml_write->pos;
     gst_ebml_write_seek (ebml, mux->duration_pos);
     gst_ebml_write_float (ebml, GST_MATROSKA_ID_DURATION,
-        (gdouble) duration / mux->time_scale);
+        gst_guint64_to_gdouble (duration / mux->time_scale));
     gst_ebml_write_seek (ebml, pos);
   }
 
@@ -1274,9 +1423,10 @@ gst_matroska_mux_best_pad (GstMatroskaMux * mux, gboolean * popped)
 
     /* if we have a buffer check if it is better then the current best one */
     if (collect_pad->buffer != NULL) {
-      if (best == NULL
-          || GST_BUFFER_TIMESTAMP (collect_pad->buffer) <
-          GST_BUFFER_TIMESTAMP (best->buffer)) {
+      if (best == NULL || !GST_BUFFER_TIMESTAMP_IS_VALID (collect_pad->buffer)
+          || (GST_BUFFER_TIMESTAMP_IS_VALID (best->buffer)
+              && GST_BUFFER_TIMESTAMP (collect_pad->buffer) <
+              GST_BUFFER_TIMESTAMP (best->buffer))) {
         best = collect_pad;
       }
     }
@@ -1285,7 +1435,6 @@ gst_matroska_mux_best_pad (GstMatroskaMux * mux, gboolean * popped)
   return best;
 }
 
-
 /**
  * gst_matroska_mux_buffer_header:
  * @track: Track context.
@@ -1298,7 +1447,7 @@ gst_matroska_mux_best_pad (GstMatroskaMux * mux, gboolean * popped)
  */
 GstBuffer *
 gst_matroska_mux_create_buffer_header (GstMatroskaTrackContext * track,
-    guint16 relative_timestamp, int flags)
+    gint16 relative_timestamp, int flags)
 {
   GstBuffer *hdr;
 
@@ -1330,19 +1479,46 @@ gst_matroska_mux_write_data (GstMatroskaMux * mux, GstMatroskaPad * collect_pad)
   GstBuffer *buf, *hdr;
   guint64 cluster, blockgroup;
   gboolean write_duration;
-  guint16 relative_timestamp;
+  gint16 relative_timestamp;
+  gint64 relative_timestamp64;
   guint64 block_duration;
+  gboolean is_video_keyframe = FALSE;
 
   /* write data */
   buf = collect_pad->buffer;
   collect_pad->buffer = NULL;
 
+  /* vorbis/theora headers are retrieved from caps and put in CodecPrivate */
+  if (collect_pad->track->xiph_headers_to_skip > 0) {
+    GST_LOG_OBJECT (collect_pad->collect.pad, "dropping streamheader buffer");
+    gst_buffer_unref (buf);
+    --collect_pad->track->xiph_headers_to_skip;
+    return GST_FLOW_OK;
+  }
+
+  /* hm, invalid timestamp (due to --to be fixed--- element upstream);
+   * this would wreak havoc with time stored in matroska file */
+  if (!GST_BUFFER_TIMESTAMP_IS_VALID (buf)) {
+    GST_WARNING_OBJECT (collect_pad->collect.pad,
+        "Invalid buffer timestamp; dropping buffer");
+    gst_buffer_unref (buf);
+    return GST_FLOW_OK;
+  }
+
   /* set the timestamp for outgoing buffers */
   ebml->timestamp = GST_BUFFER_TIMESTAMP (buf);
 
+  if (collect_pad->track->type == GST_MATROSKA_TRACK_TYPE_VIDEO &&
+      !GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DELTA_UNIT)) {
+    GST_LOG_OBJECT (mux, "have video keyframe, ts=%" GST_TIME_FORMAT,
+        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)));
+    is_video_keyframe = TRUE;
+  }
+
   if (mux->cluster) {
-    /* start a new cluster every two seconds */
-    if (mux->cluster_time + GST_SECOND * 2 < GST_BUFFER_TIMESTAMP (buf)) {
+    /* start a new cluster every two seconds or at keyframe */
+    if (mux->cluster_time + GST_SECOND * 2 < GST_BUFFER_TIMESTAMP (buf)
+        || is_video_keyframe) {
       GstMatroskaMetaSeekIndex *idx;
 
       gst_ebml_write_master_finish (ebml, mux->cluster);
@@ -1390,8 +1566,7 @@ gst_matroska_mux_write_data (GstMatroskaMux * mux, GstMatroskaPad * collect_pad)
    * for audio only files. This can be largely improved, such as doing
    * one for each keyframe or each second (for all-keyframe
    * streams), only the *first* video track. But that'll come later... */
-  if (collect_pad->track->type == GST_MATROSKA_TRACK_TYPE_VIDEO &&
-      !GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DELTA_UNIT)) {
+  if (is_video_keyframe) {
     GstMatroskaIndex *idx;
 
     if (mux->num_indexes % 32 == 0) {
@@ -1430,8 +1605,15 @@ gst_matroska_mux_write_data (GstMatroskaMux * mux, GstMatroskaPad * collect_pad)
   /* write the block, for matroska v2 use SimpleBlock if possible
    * one slice (*breath*).
    * FIXME: lacing, etc. */
-  relative_timestamp =
-      (GST_BUFFER_TIMESTAMP (buf) - mux->cluster_time) / mux->time_scale;
+  relative_timestamp64 = GST_BUFFER_TIMESTAMP (buf) - mux->cluster_time;
+  if (relative_timestamp64 >= 0) {
+    /* round the timestamp */
+    relative_timestamp64 += mux->time_scale / 2;
+  } else {
+    /* round the timestamp */
+    relative_timestamp64 -= mux->time_scale / 2;
+  }
+  relative_timestamp = relative_timestamp64 / (gint64) mux->time_scale;
   if (mux->matroska_version > 1 && !write_duration) {
     int flags =
         GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DELTA_UNIT) ? 0 : 0x80;
@@ -1540,6 +1722,9 @@ gst_matroska_mux_change_state (GstElement * element, GstStateChange transition)
       break;
     case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
       break;
+    case GST_STATE_CHANGE_PAUSED_TO_READY:
+      gst_collect_pads_stop (mux->collect);
+      break;
     default:
       break;
   }
@@ -1550,7 +1735,6 @@ gst_matroska_mux_change_state (GstElement * element, GstStateChange transition)
     case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
       break;
     case GST_STATE_CHANGE_PAUSED_TO_READY:
-      gst_collect_pads_stop (mux->collect);
       gst_matroska_mux_reset (GST_ELEMENT (mux));
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
